/*
 * Network virtual serial port.
 * Stuart Levy, slevy@ncsa.uiuc.edu, April 2000.
 *
 * Uses Unix and Win32 serial I/O code by John Stone and Mike Connell,
 * respectively; see orbserial.c.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#if WIN32
# include <winsock2.h>
  extern int close(int);

#else /*unix*/
# include <unistd.h>
# include <termios.h>
# include <sys/types.h>
# include <sys/time.h>
# if sgi
#  include <bstring.h>		/* yuck. sgi needs bstring.h for fd_set (!) */
# endif
# include <sys/socket.h>
# include <netinet/in.h>
# include <netinet/tcp.h>
# include <arpa/inet.h>
# include <netdb.h>
#endif

#include <time.h>
#include <signal.h>


#include <errno.h>

#include "orbserial.h"

OrbCommHandle ttyhandle;

#if sgi
# define HAVE_GETPTY 1
	/* sgi-specific function to allocate a pseudo-tty */
extern char *_getpty(int *masterfd, int oflag, mode_t mode, int nofork);
#endif

#if linux || __linux__
# define HAVE_GETPTY 1
# include <pty.h>	/* for openpty() */
/* extern int openpty __P ((int *__amaster, int *__aslave, char *__name,    */
/*                        struct termios *__termp, struct winsize *__winp)); */
/* _getpty emulation in terms of openpty() */
char *_getpty( int *masterfd, int oflag, mode_t mode, int junk )
{
  static char slavename[28];
  int slavefd;
  struct termios ts;
  struct winsize ws;

  ts.c_lflag = 0;
  ts.c_lflag = 0;
  ts.c_iflag = IGNBRK | IGNPAR;
  ts.c_oflag = 0;
  ts.c_cflag = CREAD | CS8 | CLOCAL | HUPCL;
  ts.c_cc[VEOL] = '\r';
  ts.c_cc[VERASE] = 0;
  ts.c_cc[VKILL] = 0;
  ts.c_cc[VMIN] = 0;
  ts.c_cc[VTIME] = 0;
  cfsetispeed(&ts, B9600);
  cfsetospeed(&ts, B9600);

  memset(&ws, 0, sizeof(ws));

  if(openpty(masterfd, &slavefd, slavename, &ts, &ws) < 0)
    return NULL;
  fchmod(slavefd, mode);
  close(slavefd);
  return slavename;
}
#endif



void tcp_nodelay( int s )
{
  static int on = 1;

#ifdef SOL_TCP
  if(setsockopt(s, SOL_TCP, TCP_NODELAY, (const char *)&on, sizeof(on)) < 0)
    perror("data socket: TCP_NODELAY");
#else /* no SOL_TCP, use SOL_SOCKET */
  if(setsockopt(s, SOL_SOCKET, TCP_NODELAY, (const char *)&on, sizeof(on)) < 0)
    perror("data socket: TCP_NODELAY");
#endif
}

#if WIN32
int inet_aton( char *str, struct in_addr *saddr )
{
  unsigned char ipaddr[4];
  int v, i;
  char *cp, *ep;

  for(cp = str, i = 0; i < 4; i++, cp = ep) {
    v = strtol(cp, &ep, 0);
    if(cp == ep || v < 0 || v > 255)
	return 0;
    ipaddr[i] = v;
    if(*ep == '.') ep++;
  }
  memcpy(saddr, ipaddr, 4);
  return 1;
}
#endif

FILE *voutf = NULL;
int nbytes, nbytesto;
char *fname = NULL;
int writedelay = 0;

void bye(int sig)
{
  fprintf(stderr, "Copied %d bytes from, %d bytes to %s, until sig %d\n",
	nbytes, nbytesto, fname, sig);
#if WIN32
  WSACleanup();
#endif
  exit(0);
}

int serial_write( OrbCommHandle ttyhandle, char *buf, int len, int writedelay )
{
  int i;
  if(writedelay == 0)
    return orb_comm_write(ttyhandle, buf, len);

  for(i = 0; i < len; i++) {
    usleep(writedelay * 1000);
    if(orb_comm_write(ttyhandle, buf+i, 1) != 1)
	break;
  }
  return i;
}

#define VERBOSELINE 10

void verboselog( FILE *outf, char *tag, char *buf, int len )
{
  int i, base;

  for(base = 0; base < len; base += VERBOSELINE) {
    if(base == 0) fprintf(outf, "%3d%s", len, tag);
    else fprintf(outf, "   %s", tag);
    for(i = 0; i < VERBOSELINE; i++) {
	int c = buf[base+i] & 0xFF;
	if(base + i >= len)
	    fprintf(outf, "   ");
	else if((c&0x7F) < ' ')
	    fprintf(outf, "%c^%c",
		(c<0x80) ? ' ' : 'M',
		(c&0x1F) + '@');
	else if(c == 0x7F)
	    fprintf(outf, " ^?");
	else if(c == 0xFF)
	    fprintf(outf, "M^?");
	else if(c >= 0x80)
	    fprintf(outf, " M%c", c&0x7F);
	else
	    fprintf(outf, "  %c", c&0x7F);
    }
    for(i = 0; i < VERBOSELINE && base+i < len; i++)
	fprintf(outf, " %02x", buf[base+i]);
    fputc('\n', outf);
  }
  fflush(outf);
}

void shipit( OrbCommHandle ttyhandle, int outfd )
{
    int n, ttyfd, maxfd, wrot;
    fd_set fds;
    char buf[BUFSIZ];
    int any = 0;

#if unix
    ttyfd = orb_comm_fd(ttyhandle);
    maxfd = ttyfd > outfd ? ttyfd : outfd;
# define Write(fd, buf, len)   write(fd, buf, len)
# define Read(fd, buf, len)    read(fd, buf, len)

#else /* WIN32 */
    ttyfd = -1;
    maxfd = outfd;
# define Write(fd, buf, len)	send(fd, buf, len, 0)
# define Read(fd, buf, len)	recv(fd, buf, len, 0)
#endif

    for(;;) {
	struct timeval nowait;
	FD_ZERO(&fds);
	if(ttyfd>=0) FD_SET((unsigned int)ttyfd, &fds);
	FD_SET((unsigned int)outfd, &fds);
	/* Under Windows, we can't select() on the serial port hardware,
	 * so just poll at .01 second intervals instead.
	 * If last poll cycle returned any data, wait 0 time instead.
	 * Under Unix, we can select() on both, so always wait indefinitely.
	 */
	nowait.tv_sec = 0;
	nowait.tv_usec = any ? 0 : 10000;
	select(maxfd+1, &fds, NULL, NULL, ttyfd>=0 ? NULL : &nowait);
	any = 0;
	if(FD_ISSET(outfd, &fds)) {
	    n = Read(outfd, buf, sizeof(buf));
	    if(n <= 0)
		break;	/* EOF from network */
	    wrot = serial_write(ttyhandle, buf, n, writedelay);
	    if(wrot != n) {
		fprintf(stderr, "writetty(%d)->%d?\n", n, wrot);
		break;
	    }
	    nbytesto += n;
	    any = 1;
	    if(voutf) verboselog( voutf, ">", buf, n ); 
	}
	if(ttyfd<0 || FD_ISSET(ttyfd, &fds)) {
	    n = orb_comm_read(ttyhandle, buf, sizeof(buf));
	    if(n > 0) {
		if((wrot = Write(outfd, buf, n)) != n) {
		    fprintf(stderr, "writenet(%d)->%d?\n", n, wrot);
		    break;
		}
		nbytes += n;
		if(voutf) verboselog( voutf, "<", buf, n ); 
	    }
	    any = 1;
	}
    }
    if(n < 0)
	perror("read");
}


int
main(int argc, char *argv[])
{
  int slisten, s;
  static int on = 1;
  struct sockaddr_in ssin;
  int portno = 0;
  char *host = NULL;
  int useptylink = 0;
  int verbose = 0;
  static char Usage[] = 
#if HAVE_GETPTY
    "Usage: ttylink [-v] [-w writems] [-t ptylinkname | devname]  [port | hostname:port | > serial-data]\n"
#else
    "Usage: ttylink  [-v] [-w writems] devname  [port | hostname:port | > serial-data]\n"
#endif
    ;

  while(argc>1 && argv[1][0] == '-' && argv[1][1] != '\0') {
    if(!strcmp(argv[1], "-v")) {
	verbose = 1;
	argc--, argv++;
    } else if(!strcmp(argv[1], "-t")) {
#if HAVE_GETPTY
	useptylink = 1;
	fname = argv[2];
	argc -= 2, argv += 2;
#else
	fprintf(stderr, "Can't do pseudo-tty (-t option) here\n");
	exit(1);
#endif
    } else if(!strcmp(argv[1], "-w")) {
	if(argc>2) writedelay = atoi(argv[2]);
	argc -= 2, argv += 2;
    } else {
	fprintf(stderr, "%s\n", Usage);
	exit(1);
    }
  }

  if(fname == NULL) {
    fname = argv[1];
    argc--, argv++;
  }

  if(argc <= 1 || (fname[0] == '-' && fname[1] != '\0')) {
    fprintf(stderr, "%s\n", Usage);
    exit(1);
  }

  if(useptylink) {
#if HAVE_GETPTY
    char *ptylink = fname;
    int fd;
    char *slavename = _getpty(&fd, O_RDWR, 0666, 0);
    if(slavename == NULL) {
	perror("ttylink: can't get pseudo-tty: ");
	exit(1);
    }
    unlink(fname);
    if(symlink(slavename, ptylink) < 0) {
	fprintf(stderr, "ttylink: can't symlink pseudo-tty to %s: ", ptylink);
	perror("");
	exit(1);
    }
    orb_comm_fdopen( fd, &ttyhandle );
#endif

  } else {
    if(orb_comm_open(fname, &ttyhandle) < 0) {
	fprintf(stderr, "ttylink: can't open %s:", fname);
	perror("");
	exit(1);
    }
  }

  if(argc > 1) {
    char *ep;
    portno = strtol(argv[1], &ep, 0);
    if(!(ep > argv[1] && *ep == '\0')) {
	portno = 0;
	ep = strchr(argv[1], ':');
	if(ep == NULL) ep = strchr(argv[1], '/');
	if(ep && *ep && ep[1]) {
	    host = argv[1];
	    *ep++ = '\0';
	    portno = strtol(ep, &ep, 0);
	    if(*ep) {
		fprintf(stderr,
		  "ttylink: expected hostname or hostname:portnumber\n");
		exit(1);
	    } 
	}
    }
  }

  memset(&ssin, 0, sizeof(ssin));
  ssin.sin_family = AF_INET;
  ssin.sin_port = htons( (unsigned short) portno );
  ssin.sin_addr.s_addr = INADDR_ANY;
  if(host) {
    struct hostent *hp;
    if(inet_aton(host, &ssin.sin_addr)) {
	/* fine */
    } else if((hp = gethostbyname(host)) != NULL) {
	memcpy(&ssin.sin_addr, hp->h_addr_list[0], sizeof(ssin.sin_addr));
    } else {
	fprintf(stderr, "ttylink: %s: unknown hostname\n", host);
	exit(1);
    }
  }


  signal(SIGINT, bye);
  signal(SIGTERM, bye);
#ifdef SIGHUP
  signal(SIGHUP, bye);	/* not under WIN32 */
#endif

#if WIN32
  { struct WSAData wsad;
    WSAStartup( MAKEWORD(1, 1), &wsad );
  }
#endif

  nbytes = nbytesto = 0;

  if(verbose)
    voutf = portno ? stdout : stderr;

  if(portno) {
    s = socket(AF_INET, SOCK_STREAM, 0);
    if(s<0) {
	perror("socket");
	exit(1);
    }

    if(host) {
	if(connect(s, (struct sockaddr *)&ssin, sizeof(ssin)) < 0) {
	    perror("connect");
	    exit(1);
	}
	tcp_nodelay( s );
	shipit( ttyhandle, s );

    } else {
	slisten = s;
	if(setsockopt(slisten, SOL_SOCKET, SO_REUSEADDR, (const char *)&on, sizeof(on)) < 0)
	    perror("slisten: setsockopt");

	if(bind(slisten, (struct sockaddr *)&ssin, sizeof(ssin)) < 0) {
	    perror("giving up: bind");
	    exit(1);
	}

	listen(slisten, 1);

	for(;;) {
	    fd_set fds;
	    struct sockaddr_in from;
	    int fromlen = sizeof(from);

	    FD_ZERO(&fds);
	    FD_SET((unsigned int)slisten, &fds);
	    if(select(slisten+1, &fds, NULL, NULL, NULL) <= 0 && errno == EINTR)
		continue;
	    s = accept(slisten, (struct sockaddr *)&from, &fromlen);
	    if(s < 0) {
		perror("accept");
		exit(1);
	    }
	    fprintf(stderr, "connect from %s\n", inet_ntoa(from.sin_addr));
	    tcp_nodelay( s );
	    shipit( ttyhandle, s );
	    close(s);
	    fprintf(stderr, "ttylink: EOF after %d bytes from, %d bytes to %s\n", nbytes, nbytesto, fname);
	}
    }
  } else {
	/* If no port, just write to stdout */
    shipit( ttyhandle, 1 );
  }
  bye(0);
  return 0;
}
